import os, time, re, email, email.Utils, mailbox
try: import cPickle as pickle
except: import pickle

# ------------------------------------------------------------ mail utilities
def linesplit(text):
    if text[-1:] == '\n': return text[:-1].split('\n')
    else: return text.split('\n')

def linejoin(lines):
    return '\n'.join(lines) + '\n'

isodatepat = re.compile(
    r'(\d{4})-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d) *([+-]\d+)?')
def unixdate(date):
    match = isodatepat.match(date.strip())
    if match:
        y, l, d, h, m, s = map(int, match.group(1, 2, 3, 4, 5, 6))
        tzo = match.group(7) or 0
        if tzo: tzo = int(tzo[:-2]) * 3600 + int(tzo[-2:]) * 60
    else:
        result = email.Utils.parsedate_tz(date)
        if not result: raise ValueError, 'unrecognizable date %r' % date
        y, l, d, h, m, s, _, _, _, tzo = result
    return time.mktime((y, l, d, h, m, s, 0, 0, 0)) - time.timezone - tzo

def isodate(date):
    if type(date) is str: date = unixdate(date)
    y, l, d, h, m, s, _, _, _ = time.gmtime(date)
    return '%04d-%02d-%02d %02d:%02d:%02d UTC' % (y, l, d, h, m, s)

months = '. Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec'.split()
def simpledate(date):
    dt = unixdate(date)
    now = time.time()
    diff = abs(now - dt)
    direction = now > dt and ' ago' or ' from now'
    if diff < 60*2: return '%d s' % diff + direction
    elif diff < 60*60*2: return '%d m' % (diff/60) + direction
    elif diff < 24*60*60*2: return '%d h' % (diff/60/60) + direction
    elif diff < 30*24*60*60: return '%d d' % (diff/60/60/24) + direction
    else:
        y, l, d, h, m, s, _, _, _ = time.localtime(dt)
        if y == time.localtime(now)[0]: return '%s %d' % (months[l], d)
        else: return '%d %s %d' % (y, months[l], d)

replypat = re.compile(r'[rR][eE]:\s*')
def normsubject(subject, prefix=' '):
    decoded = email.Header.decode_header(subject)
    if isinstance(decoded, list):
        parts = []
        for part, charset in decoded:
            if charset:
                parts.append(part.decode(charset))
            else:
                parts.append(part)
        decoded = u''.join(parts)
    subject = decoded.encode('iso-8859-1').strip()
    while 1:
        if subject.startswith(prefix):
            subject = subject[len(prefix):].strip()
        elif replypat.match(subject):
            subject = subject[replypat.match(subject).end():].strip()
        else:
            break
    words = subject.split()
    length = len(' '.join(words))
    while length > 60 and length - len(words[-1]) > 30:
        length -= len(words[-1]) + 1
        words.pop()
    return ' '.join(words)

idpat = re.compile('<(\S+)>')
def extractid(header):
    match = idpat.search(header)
    if match:
        return match.group(1)

def splitfooter(lines, footerlines):
    if not footerlines or len(lines) < len(footerlines):
        return lines, []
    for i in range(-len(footerlines), 0):
        if (footerlines[i] is not None and 
            footerlines[i].strip() != lines[i].strip()):
            return lines, []
    else:
        return lines[:-len(footerlines)], lines[-len(footerlines):]

signaturepats = [(re.compile(r'^-- $'), 20),
                 (re.compile(r'^[_=-]{2,}\s*$'), 10),
                 (re.compile(r'^--.{,30}'), 5)]
def splitsig(lines):
    last = len(lines) - 1
    while last > 0 and not lines[last].strip(): last -= 1
    for pattern, siglines in signaturepats:
        for i in range(last, max(-1, last - siglines), -1):
            if pattern.match(lines[i]):
                return lines[:i], lines[i:]
    return lines, []

refheaderpat = re.compile(r'^\s*(>>>>+.*|On \d+ \w+ \d+, *[\w ]+ *\w+:\s*)$')
def isrefheader(line):
    return refheaderpat.match(line) is not None

refheaderattrpat = re.compile(r'^\s*>>>>+.*"([^"]+)"\s*=')
def refheaderattr(line):
    match = refheaderattrpat.match(line)
    if match: return match.group(1)

quotepat = re.compile(r'^( ?(?!>>>)(>>?| ?--\])|    [A-Z]\w*> )')
def unquote(lines, attrs=None):
    pat = attrs and re.compile(r'^\s*(>|%s)>' % '|'.join(attrs)) or quotepat
    unquoted = []
    for line in lines:
        match = pat.match(line)
        if match: line = line[match.end():]
        unquoted.append(line)
    return unquoted

def isquoted(line, attrs=None):
    pat = attrs and re.compile(r'^\s*(>|%s)>' % '|'.join(attrs)) or quotepat
    return pat.match(line) is not None

endpointpats = [re.compile(r'.*[.!?]["\']?($|  )'),
                re.compile(r'.*[.!?]["\']? '),
                re.compile(r'.*:\s*$'),
                re.compile(r'.*(?=\s[([])'),
                re.compile(r'.*[\w\s](?=--[\w\s])'),
                re.compile(r'.*,["\']? ')]
def getendpoint(lines, maxlines=2):
    lnum = 0
    while lnum + 1 < len(lines) and not lines[lnum].strip(): lnum += 1
    start = lnum
    checklines = range(start, min(start + maxlines, len(lines)))
    checklines.reverse()
    for pattern in endpointpats:
        for lnum in checklines:
            match = pattern.match(lines[lnum])
            if match:
                return start, lnum, match.end()
    lnum = min(start + maxlines, len(lines)) - 1
    return start, lnum, len(lines[lnum].rstrip())

def getpoint(lines, maxlines=2):
    start, lnum, pos = getendpoint(lines, maxlines)
    point = ' '.join(lines[start:lnum] + [lines[lnum][:pos]])
    return ' '.join(point.split()), (start, 0), (lnum, pos)

# ------------------------------------------------------------ HTML utilities
def plural(number, suffix='s'):
    return (number != 1) and suffix or ''

def flatten(stuff):
    if type(stuff) in [list, tuple]: return ''.join(map(flatten, stuff))
    else: return str(stuff)

tagpat = re.compile(r'<[^>]*>')
def striphtml(text):
    return re.sub(tagpat, '', text).replace(
        '&lt;', '<').replace('&gt;', '>').replace('&amp;', '&')

def escape(text):
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')

import pydoc
markup = pydoc.html.markup

def formatpoint(point):
    return re.sub(r'\*(\w+\s?\w*\s?\w*)\*', r'<em>\1</em>',
                  re.sub(r'\*\*(\w+\s?\w*\s?\w*)\*\*', r'<strong>\1</strong>',
                         markup(point)))

_rename={'pad': 'cellpadding', 'space': 'cellspacing', 'c': 'class'}.get
def _attrs(d):
    return ''.join([' %s="%s"' % (_rename(k, k), d[k])
                    for k in d if d[k] is not None])

def _mktag(name): return lambda *stuff, **attrs: [
    '<%s%s\n>' % (name, _attrs(attrs)), stuff, '</%s>' % name]
def _mkempty(name): return lambda **attrs: '<%s%s>' % (name, _attrs(attrs))

def _tagdef(tag, **defaults):
    def newtag(*stuff, **attrs):
        dict = defaults.copy()
        dict.update(attrs)
        return tag(*stuff, **dict)
    return newtag

p, br, nbsp = '\n<p>', '<br>\n', '&nbsp;'
for t in 'html head title frameset body pre a table tr td th'.split():
    globals()[t] = _mktag(t)
for t in 'meta img frame'.split(): globals()[t] = _mkempty(t)
table = _tagdef(table, pad=0, space=0, border=0)
tdc = _tagdef(td, align='center')
tdr = _tagdef(td, align='right')

def span(*stuff, **attrs):
    c = attrs.get('c')
    start = '<span'
    if c: start += ' class="%s"' % c
    attrs['c'] = None
    start += _attrs(attrs) + '>'
    end = '</span>'
    if c: end += '<!--%s-->' % c
    return [start, stuff, end]

def findspan(text, cls, id=None):
    tag = '<span class="%s"%s' % (cls, id and (' id="%s"' % id) or '')
    start = text.find('>', text.find(tag)) + 1
    end = text.find('</span><!--%s-->' % cls, start)
    return start, end

def getspan(text, cls, id=None):
    start, end = findspan(text, cls, id)
    return text[start:end]

def write(filename, *stuff):
    file = open(filename, 'w')
    file.write(flatten(stuff))
    file.close()
    print 'wrote', filename

# ------------------------------------------------------------------ database
criticonpat = re.compile(r'\s*((\[[-+?#]?\])?)')
sectypes = {'[+]': 'support', '[-]': 'issue', '[]': None,
            '[?]': 'question', '[#]': 'comment', '': None}
secicons = {'support': 'aff.png', 'issue': 'neg.png',
            'question': 'query.png', 'comment': 'info.png', None: 'top.png'}

votepat = re.compile(r'\s*([+-][01])')
def addcriticon(point):
    match = votepat.match(point)
    if match:
        vote = match.group(1)
        return {'+0': '[#] ', '-0': '[-] ',
                '+1': '[+] ', '-1': '[-] '}[vote] + point
    return point

class UnixMailbox(mailbox.PortableUnixMailbox):
    _fromlinepattern = r'From (\w+@|.*<\w+@)'

class DB:
    def __init__(self, filename=None):
        if filename:
            import marshal
            file = open(filename)
            dict = marshal.load(file)
            file.close()
            for key in dict: self.__dict__[key] = dict[key]
        else:
            self.nmsgs = 0
            self.nsecs = []
            self.nauthors = 0
            self.msgindex = {}
            self.authorindex = {}
            self.subjectindex = {}
            self.weakparent = {}
            for key in ('msgid msgparent msgkids msgauthor msgdate msgsubject '
                        'sectype secpoint secparent seckids '
                        'authoraddr authorname authornick').split():
                self.__dict__[key] = []

    def __repr__(self):
        return '<Archiver.DB with %d message%s, %d author%s>' % (
            self.nmsgs, plural(self.nmsgs),
            self.nauthors, plural(self.nauthors))

    def dump(self, filename):
        import marshal
        file = open(filename, 'w')
        marshal.dump(self.__dict__, file)
        file.close()

    def author(self, address, name=None, nick=None):
        if address in self.authorindex: return self.authorindex[address]
        if name is None: raise KeyError, 'address %r not found' % address
        assert len(self.authoraddr) == len(self.authorname) == self.nauthors
        authorindex = self.nauthors
        self.authorindex[address] = authorindex
        self.authoraddr.append(address)
        self.authorname.append(name)
        if not nick: nick = address.split('@')[0]
        self.authornick.append(nick)
        self.nauthors += 1
        return authorindex

    def addmsg(self, msgid, authorindex, date, subject):
        for list in [self.msgid, self.msgparent, self.msgkids,
                     self.msgauthor, self.msgdate, self.msgsubject,
                     self.sectype, self.secpoint, self.secparent, self.seckids]:
            assert len(list) == self.nmsgs
        msgi = self.nmsgs
        self.msgindex[msgid] = msgi
        self.msgid.append(msgid)
        self.msgparent.append(None)
        self.msgkids.append([])
        self.msgauthor.append(authorindex)
        self.msgdate.append(isodate(date))
        self.msgsubject.append(subject)
        self.nsecs.append(0)
        self.sectype.append([])
        self.secpoint.append([])
        self.secparent.append([])
        self.seckids.append([])
        self.subjectindex.setdefault(subject, []).append(msgi)
        self.nmsgs += 1
        return msgi

    def msgreply(self, msgi, replymsgi):
        print 'msg %d -> %d' % (msgi, replymsgi)
        self.msgparent[replymsgi] = msgi
        self.msgkids[msgi].append(replymsgi)

    def addsec(self, msgi, type, point):
        if type not in sectypes.values():
            raise ValueError, 'invalid type %r' % type
        for list in [self.sectype[msgi], self.secpoint[msgi],
                     self.secparent[msgi], self.seckids[msgi]]:
            assert len(list) == self.nsecs[msgi]
        seci = self.nsecs[msgi]
        self.sectype[msgi].append(type)
        self.secpoint[msgi].append(point)
        self.secparent[msgi].append(None)
        self.seckids[msgi].append([])
        self.nsecs[msgi] += 1
        return seci

    def secreply(self, (msgi, seci), (replymsgi, replyseci)):
        index = (msgi, seci)
        replyindex = (replymsgi, replyseci)
        if replyindex in self.seckids[msgi][seci]: return
        print 'sec %r -> %r' % (index, replyindex)
        self.secparent[replymsgi][replyseci] = index
        self.seckids[msgi][seci].append(replyindex)

MARK = '\xff\xff'

class Archive:
    DBFILE = 'archive.db'
    MESSAGEFILE = 'msg%05d.html'
    INDEXFILE = 'index.html'
    THREADFILE = 'thr%05d.html'
    OUTLINEFILE = 'out%05d.html'
    LISTINGFILE = 'listing.html'

    DOCTYPE = '<!doctype html public "-//W3C//DTD HTML 4.0//EN">'
    FRAMEDOCTYPE = '<!doctype html public "-//W3C DTD HTML 4.0 Frameset//EN">'
    # d.msg holds msgid of message currently displayed in message frame
    # d.himsg holds msgid of message currently highlighted in outline frame
    # d.hinsecs holds nsecs of message currently highlighted in outline frame
    # d.hisec holds secid of section currently highlighted in outline frame
    HEADTAGS = '''
<link rel="stylesheet" href="style.css">
<script type="text/javascript"><!--

dom = document.documentElement;
d = parent.document;

function set(doc, id, cn) {
    var e = doc.getElementById(id);
    if (e) e.className = cn;
}
function setboth(msg, sec, cn) {
    var id = 'msg' + msg + '-sec' + sec;
    set(parent.frames.outline.document, id, cn);
    set(parent.frames.message.document, id, cn);
}
function setall(msg, count, cn) {
    for (var sec=0; sec<count; sec++) {
        var id = 'msg' + msg + '-sec' + sec;
        set(parent.frames.outline.document, id, cn);
    }
}
function load(msg) {
    if (dom) {
        if (d.himsg == msg) {
            setboth(d.himsg, d.hisec, 'sel fresh');
        }
        d.msg = msg;
    }
}
function gothread(thr) {
    if (dom) {
        set(document, 'thr' + document.thr, 'record');
        document.thr = thr
        set(document, 'thr' + document.thr, 'sel record');
    }
}
function go(msg, nsecs, sec) {
    var id = 'msg' + msg + '-sec' + sec;
    if (dom) {
        if (parent.frames.outline.document.getElementById(id)) {
            setall(d.himsg, d.hinsecs, 'fresh');
            setall(msg, nsecs, 'msgsel fresh');
            setboth(d.himsg, d.hisec, 'fresh');
            setboth(msg, sec, 'sel fresh');
            d.himsg = msg;
            d.hinsecs = nsecs;
            d.hisec = sec;
        }
    }
} //-->

</script>'''

    def __init__(self, dir, listname='Mailman', subjectprefix=' ',
                            footerlines=[], maxsiglines=5):
        self.dir = dir
        self.listname = listname
        self.subjectprefix = subjectprefix
        self.footerlines = footerlines[:]
        self.maxsiglines = maxsiglines
        self.dbfilename = os.path.join(dir, self.DBFILE)
        if os.path.exists(self.dbfilename):
            file = open(self.dbfilename)
            self.db = pickle.load(file)
            file.close()
        else: self.db = DB()

    def __repr__(self):
        return '<Archiver.Archive for %r at %s>' % (self.listname, self.dir)

    def gettext(self, msgi, seci=None, point=0):
        filename = os.path.join(self.dir, self.MESSAGEFILE % msgi)
        if not os.path.exists(filename): return ''
        file = open(filename)
        text = file.read()
        file.close()
        msg = getspan(text, 'body')
        if seci is None: return striphtml(msg)
        sec = getspan(msg, 'fresh', 'msg%d-sec%d' % (msgi, seci))
        if not point: return striphtml(sec)
        point = getspan(sec, 'point')
        return striphtml(point)

    def findquote(self, quotelines, attrs, msgi=None, subject=None):
        i = 0
        while i < len(quotelines) and not (
            isquoted(quotelines[i]) or isquoted(quotelines[i], attrs)): i += 1
        quotelines = unquote(quotelines[i:], attrs)
        i = len(quotelines) - 1
        while i > 0 and not quotelines[i].strip(): i -= 1
        while i > 0 and quotelines[i].strip(): i -= 1
        quotewords = ' '.join(quotelines[i:]).split()[:12]
        checklist = []
        if msgi is not None:
            while self.db.msgparent[msgi]:
                msgi = self.db.msgparent[msgi]
                checklist.append(msgi)
        if subject is not None:
            indices = self.db.subjectindex.get(subject, [])[:]
            indices.reverse()
            for msgi in indices:
                if msgi not in checklist: checklist.append(msgi)
        indices = range(len(self.db.msgid))
        indices.reverse()
        for msgi in indices:
            if msgi not in checklist: checklist.append(msgi)

        for msgi in checklist:
            for seci in range(len(self.db.sectype[msgi])):
                sec = self.gettext(msgi, seci).split()
                for i in range(len(sec)):
                    if sec[i:i+len(quotewords)] == quotewords:
                        return msgi, seci

    def flush(self):
        file = open(os.path.join(self.dir, self.DBFILE), 'w')
        pickle.dump(self.db, file)
        file.close()

    def slurp(self, mailboxfile):
        box = UnixMailbox(mailboxfile, email.message_from_file)
        for message in box:
            self.add(message)

    def fixparent(self, msgi, seci):
        icon = img(src='reply.png', border=0)
        replies = []
        seen = {}
        for kmsgi, kseci in self.db.seckids[msgi][seci]:
            if kmsgi in seen: continue
            seen[kmsgi] = 1
            nick = self.db.authornick[self.db.msgauthor[kmsgi]]
            id = 'msg%d-sec%d' % (kmsgi, kseci)
            onhover = 'go(%d, %d, %d)' % (kmsgi, self.db.nsecs[kmsgi], kseci)
            escpoint = self.db.secpoint[kmsgi][kseci].replace('"', '&quot;')
            reply = [' ', a(span(nick, c='nick'), icon,
                            href=self.MESSAGEFILE % kmsgi + '#' + id,
                            title=escpoint, onmouseover=onhover)]
            replies.append(reply)
        filename = os.path.join(self.dir, self.MESSAGEFILE % msgi)
        file = open(filename)
        text = file.read()
        file.close()
        id = 'msg%d-sec%d-reply' % (msgi, seci)
        start, end = findspan(text, 'reply', id)
        text = text[:start] + flatten(replies) + text[end:]
        file = open(filename, 'w')
        file.write(text)
        file.close()

    def add(self, message):
        # Examine the headers and add a message record to the database.
        db = self.db
        name, address = email.Utils.parseaddr(message['from'])
        authorindex = db.author(address, name)
        date = isodate(message['date'])
        subject = normsubject(message['subject'], self.subjectprefix)
        id = (extractid(message['message-id'] or '') or
              '_'.join([address] + date.split()))
        if id in db.msgindex: return
        msgi = db.addmsg(id, authorindex, date, subject)
        print '\n== %d <%s>' % (msgi, id)
        parent = db.msgindex.get(extractid(message['in-reply-to'] or ''))
        if parent: db.msgreply(parent, msgi)

        # Get the text of the message and separate out the main body.
        text = message.get_payload(decode=1)
        while text == None:
            part = message
            for subpart in part.get_payload():
                if subpart.get_type() == 'text/plain':
                    part, text = subpart, subpart.get_payload(decode=1)
                    break
        lines = linesplit(text)
        while lines and not lines[0].strip(): lines.pop(0)
        while lines and not lines[-1].strip(): lines.pop()
        lines, footer = splitfooter(lines, self.footerlines)
        lines, signature = splitsig(lines)

        # Split the body of the message into quoted and fresh sections.
        sec, secs, mode, lastblank, attrs = [], [], 'fresh', 0, []
        for line in lines:
            nextmode = mode
            if isquoted(line, attrs):
                nextmode = 'quote'
            elif isrefheader(line):
                nextmode = 'quote'
                attr = refheaderattr(line)
                if attr: attrs.append(attr)
            elif line.strip():
                if mode == 'quote':
                    if line.lstrip() == line and \
                        (len(line.strip()) < 15 or
                        (len(line.strip()) > 75 and ' ' not in line.strip())):
                        lastline = unquote(sec[-1:])[0].rstrip()
                        extraword = line.strip()
                        n = len(lastline)
                        print 'check', repr(lastline), repr(extraword), n
                        print 40, n, n + 1 + len(extraword), 79
                        if 40 <= n < n + 1 + len(extraword) or \
                           len(extraword) >= 79:
                            sec[-1] = sec[-1].rstrip() + ' ' + extraword
                            lastblank = 0
                            print 'fixup', sec
                            continue
                nextmode = 'fresh'
            if lastblank and line.strip() or nextmode != mode:
                if sec: secs.append((mode, sec, attrs))
                mode, sec = nextmode, []
            sec.append(line)
            lastblank = not line.strip()
        if sec: secs.append((mode, sec, attrs))

        # Aggregate sections according to trailing punctuation or indentation.
        i = 1
        while i < len(secs):
            prevmode, prevsec, prevattrs = secs[i - 1]
            nextmode, nextsec, nextattrs = secs[i]
            #indented = 0 not in [line[:1] in [' ', '\t'] for line in nextsec]
            tail = ''.join(prevsec[-3:]).rstrip()[-1:]
            #if (indented or tail == ':' or
            #    tail in ['-', ','] and nextmode == 'fresh'):
            if tail in [':', ']'] and nextmode == 'quote':
                secs[i-1:i+1] = [(nextmode, prevsec + nextsec, nextattrs)]
            else: i += 1

        if signature: secs.append(('sig', signature, []))
        if footer: secs.append(('footer', footer, []))

        # Register fresh sections in the database and mark up the message body.
        mbody = []
        lastquote = connected = weaklink = None
        seci = None
        for mode, sec, attrs in secs:
            smsgi = None
            if mode == 'fresh':
                sec[0] = addcriticon(sec[0])
                match = criticonpat.match(sec[0])
                criticon = match.group(1)
                type = sectypes[criticon]
                sec[0] = sec[0][len(criticon):]
                point, (sline, spos), (eline, epos) = getpoint(sec)
                smsgi, seci = msgi, db.addsec(msgi, type, point)
                if lastquote:
                    if weaklink:
                        db.weakparent[msgi, seci] = lastquote
                    else:
                        if lastquote in db.weakparent:
                            db.secreply(db.weakparent[lastquote], lastquote)
                            self.fixparent(*db.weakparent[lastquote])
                        db.secreply(lastquote, (msgi, seci))
                        self.fixparent(*lastquote)
                        connected = 1
                    weaklink = 1
                    # lastquote = None

                sec[eline] = sec[eline][:epos] + MARK + sec[eline][epos:]
                sec[sline] = sec[sline][:spos] + MARK + sec[sline][spos:]
                sec = markup(linejoin(sec))
                before, point, after = sec.split(MARK)
                stripped = point.rstrip()
                if len(point) != len(stripped):
                    after = point[len(stripped):] + after
                sec = [criticon, before, span(stripped, c='point'), ' ', after]
            else:
                if mode == 'quote':
                    quoteloc = self.findquote(sec, attrs, parent, subject)
                    if quoteloc: smsgi, seci = quoteloc
                    lastquote, weaklink = quoteloc, None
                sec = markup(linejoin(sec))
            sec = flatten(sec)
            stripped = sec.rstrip()
            sec, rest = stripped, sec[len(stripped):]
            if smsgi is not None:
                secid = 'msg%d-sec%d' % (smsgi, seci)
                sec = span(sec, c=mode, id=secid)
            sec = span(sec, c=mode)
            if smsgi is not None:
                onhover = 'go(%d, %d, %d)' % (smsgi, db.nsecs[smsgi], seci)
                if smsgi == msgi:
                    onclick = "parent.frames.outline.location=parent.document.outlinefile + '#%s'" % secid
                else:
                    onclick = "parent.frames.message.location='%s'" % (
                        self.MESSAGEFILE % smsgi + '#' + secid)
                sec = a(sec, onmouseover=onhover, onclick=onclick)
            mbody.append(sec)
            if mode == 'fresh':
                mbody.append(span('', c='reply', id=secid + '-reply'))
            mbody.append(rest)

        if parent and not connected and seci is not None:
            db.secreply((parent, 0), (msgi, 0))
            self.fixparent(parent, 0)

        # Generate the page.
        headers = [[span([span(h.capitalize() + ': ', c='headername'), '\n',
                          markup(message[h])], c=h), br]
                   for h in 'from date subject'.split()]
        if parent:
            ai = db.msgauthor[parent]
            resp = '%s <%s>' % (db.authorname[ai], db.authoraddr[ai])
            headers[1:1] = [span([span('Replying To: ', c='headername'), '\n',
                                  markup(resp)], c='replyto'), br]

        onload = 'load(%d)' % msgi

        write(os.path.join(self.dir, self.MESSAGEFILE % msgi), self.DOCTYPE,
              html(head(title(self.listname, ': ', subject),
                        meta(name='id', content=id),
                        meta(name='date', content=date)), self.HEADTAGS),
                   body(p, span(headers, c='header'), '\n\n',
                        pre(span(mbody, c='body')),
                        onload=onload, c='message', id='msg%d' % msgi))

    def getthreads(self, msgi, seci, level=0):
        db = self.db
        print 'seen', (msgi, seci)
        self.seen[(msgi, seci)] = 1
        self.threadmessages[msgi] = 1
        agesecs = unixdate(db.msgdate[-1]) - unixdate(db.msgdate[msgi])
        age = min(int(agesecs / 3600 / 24), 7)
        if (db.seckids[msgi][seci] or
            db.secparent[msgi][seci] or
            seci == 0 and msgi not in self.seen):
            self.seen[msgi] = 1
            id = 'msg%d-sec%d' % (msgi, seci)
            icon = img(src=secicons[db.sectype[msgi][seci]])
            onhover = "go(%d, %d, %d)" % (msgi, db.nsecs[msgi], seci)
            point = span(formatpoint(db.secpoint[msgi][seci]), c='point')
            item = [a(span(db.authornick[db.msgauthor[msgi]], c='nick'),
                      href=self.MESSAGEFILE % msgi + '#' + id,
                      target='message', onmouseover=onhover), ': ', point]
            self.threads += table(tr(
                td(nbsp * 4 * level),
                td(span(icon, nbsp, item, c='age%d' % age),
                   width='100%', c='fresh', id=id)),
                width='100%')
        for (msgi, seci) in db.seckids[msgi][seci]:
            self.getthreads(msgi, seci, level + 1)

    def thread(self, msgi, thri, threads):
        t = self.listname + ': ' + self.db.msgsubject[msgi]
        write(os.path.join(self.dir, self.OUTLINEFILE % thri),
              self.DOCTYPE,
              html(head(self.HEADTAGS),
                   body(self.threads, c='outline')))
        write(os.path.join(self.dir, self.THREADFILE % thri),
              self.FRAMEDOCTYPE,
              html(head(title(t),
                        "<script>document.outlinefile='%s'</script>" %
                        (self.OUTLINEFILE % thri)),
                   frameset(frame(src=self.OUTLINEFILE % thri, name='outline'),
                            frame(src=self.MESSAGEFILE % msgi, name='message'),
                            cols='30%,*', border=2, topmargin=0, leftmargin=0)))

    def outline(self):
        self.seen = {}
        db = self.db
        thri = 0
        tlist = []
        for msgi in range(db.nmsgs):
            self.threadmessages = {}
            self.threads = []
            root = msgi
            for seci in range(db.nsecs[msgi]):
                if (msgi, seci) not in self.seen:
                    self.getthreads(msgi, seci)
            while 1:
                didstuff = 0
                for msgi in self.threadmessages.keys():
                    for seci in range(db.nsecs[msgi]):
                        if (msgi, seci) not in self.seen:
                            didstuff = 1
                            self.getthreads(msgi, seci)
                if not didstuff: break
            if self.threads:
                self.thread(root, thri, self.threads)
                msgs = self.threadmessages.keys()
                dates = [db.msgdate[i] for i in msgs]
                tlist.append(
                    (max(dates), len(msgs), thri, root, min(dates)))
                thri += 1
        tlist.sort()
        tlist.reverse()
        sp = td(nbsp);
        rows = [tr(th('first', c='date'), sp, th('last', c='date'), sp,
                   th('msgs', align='right', c='count'), sp,
                   th(span('starter:', c='nick'), align='right', c='from'), sp,
                   th('topic', align='left', c='subject', width='70%'),
                   c='heading')]
        for lastdate, messages, thri, msgi, firstdate in tlist:
            first, last = simpledate(firstdate), simpledate(lastdate)
            nick = db.authornick[db.msgauthor[msgi]]
            topic = span(db.msgsubject[msgi], c='subject')
            onclick = 'gothread(%d)' % thri
            rows.append(tr(tdc(first, c='date'), sp, tdc(last, c='date'), sp,
                           tdr(messages, c='count'), sp,
                           tdr(span(nick, ':', c='nick'), c='from'), sp,
                           td(a(topic, href=self.THREADFILE % thri,
                                onclick=onclick, target='thread')),
                           id='thr%d' % thri, c='record', valign='baseline'))
        write(os.path.join(self.dir, self.LISTINGFILE),
              self.DOCTYPE,
              html(head(self.HEADTAGS),
                   body(table(rows, width='100%'), c='listing')))
        write(os.path.join(self.dir, 'index.html'),
              self.FRAMEDOCTYPE,
              html(head(title(self.listname)),
                   frameset(frame(src=self.LISTINGFILE, name='listing'),
                            frame(src=self.THREADFILE % 0, name='thread'),
                            rows='20%,*', border=2, topmargin=0, leftmargin=0)))
